/*
 *  fcnc4m.c
 * 
 *  userspace HAL program for FCNC4M panel from Adtech.
 * 
 *  Copyright (C) 2019 Kjell Claesson <kjell.claesson at gmail.com>.
 * 
 *  Based on a work (test-modbus program, part of libmodbus) which is
 *  Copyright (C) 2001-2005 Stéphane Raimbault <stephane.raimbault@free.fr>
 * 
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation, version 2.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301-1307  USA.
 * 
 *  see 'man fcnc4m'.
 * 
 */


#include "fcnc4m.h"

void  windup(param_pointer p) 
{

    if (p->haldata && *(p->haldata->errorcount)) {
        fprintf(stderr,"%s: %d modbus errors\n",p->progname, *(p->haldata->errorcount));
        fprintf(stderr,"%s: last command register: 0x%.4x\n",p->progname, p->failed_reg);
        fprintf(stderr,"%s: last error: %s\n",p->progname, modbus_strerror(p->last_errno));
    }

    if (p->hal_comp_id >= 0)
        hal_exit(p->hal_comp_id);
    if (p->ctx)
        modbus_close(p->ctx);
}

static void toggle_modbus_debug(int sig)
{
    param.modbus_debug = !param.modbus_debug;
    modbus_set_debug(param.ctx, param.modbus_debug);
}

static void toggle_debug(int sig)
{
    param.debug = !param.debug;
}

static void quit(int sig) 
{
    if (param.debug)
        fprintf(stderr,"quit(connection_state=%d)\n",connection_state);
    
    switch (connection_state) {
        
        case CONNECTING:  
            // modbus_tcp_accept() or TCP modbus_connect()  were interrupted
            // these wont return to the main loop, so exit here
            windup(&param);
            exit(0);
            break;
            
        default:
            connection_state = DONE;
            break;
    }
}


void usage(int argc, char **argv) {
    printf("Usage:  %s [options]\n", argv[0]);
    printf("This is a userspace HAL program, typically loaded using the halcmd \"loadusr\" command:\n"
    "    loadusr fcnc4m [options]\n"
    "Options are:\n"
    "-P or --port <port-name> (default /dev/ttyS0)\n"
    "    Set communicationport to use.\n"
    "-d or --debug\n"
    "    Turn on debugging messages. Toggled by USR1 signal.\n"
    "-m or --modbus-debug\n"
    "    Turn on modbus debugging.  This will cause all modbus messages\n"
    "    to be printed in hex on the terminal. Toggled by USR2 signal.\n");
}

int write_data(modbus_t *ctx, haldata_t *haldata, param_pointer p)
{
    bool_t send = FALSE;
    uint16_t tmp = 0;
    const uint16_t *cmd1_reg = &tmp;
    
    
    if (!(*haldata->x_is_home == p->x_home)) {
        tmp = *haldata->x_is_home > 0? 0x0201 : 0x0001;
        send = TRUE;
        p->x_home = *haldata->x_is_home;
    }
    
    if (!(*haldata->y_is_home == p->y_home) && !send) {
        tmp = *haldata->y_is_home > 0? 0x0202 : 0x0002;
        send = TRUE;
        p->y_home = *haldata->y_is_home;
    }
    
    if (!(*haldata->z_is_home == p->z_home) && !send) {
        tmp = *haldata->z_is_home > 0? 0x0203 : 0x0003;
        send = TRUE;
        p->z_home = *haldata->z_is_home;
    }
    
    if (send == TRUE){
        DBG("write_data: cmd1_reg=0x%4.4X ", cmd1_reg);
        
        if (modbus_write_registers(ctx, REG_LED, 1, cmd1_reg) < 0) {
            p->failed_reg = REG_LED;
            (*haldata->errorcount)++;
            p->last_errno = errno;
            return errno;
        }
        send = FALSE;
    }
    
    return 0;
}


#define GETREG(reg,into)					\
do {							\
    curr_reg = reg;						\
    if (modbus_read_registers(ctx, reg, 1, into) != 1)	\
        goto failed;					\
} while (0)


int read_data(modbus_t *ctx, haldata_t *haldata, param_pointer p)
{
    int retval;
    uint16_t curr_reg;
    
    
    GETREG(SR_KNOBS_SRE, &curr_reg);
    *(haldata->e_stop) = curr_reg & 0x0001;
    *(haldata->stop) = curr_reg & 0x0002;
    *(haldata->run) = curr_reg & 0x0004;
    *(haldata->mode) = curr_reg >> 4 & 0x000f;
    *(haldata->speed_override) = curr_reg >> 8 & 0x000f;
    *(haldata->feed_override) = curr_reg >> 12 & 0x000f;
    
    GETREG(SR_KEYBOARD, &curr_reg);
    *(haldata->first_key) = curr_reg & 0x00ff;
    *(haldata->second_key) = curr_reg >> 8 & 0x00ff;
    
    
    p->last_errno = retval = 0;
    return 0;
    
    failed:
    p->failed_reg = curr_reg;
    p->last_errno = errno;
    (*haldata->errorcount)++;
    if (p->debug)
        fprintf(stderr, "%s: read_data: modbus_read_registers(0x%4.4x): %s\n",
                p->progname, curr_reg, modbus_strerror(errno));
        return p->last_errno;
}

#undef GETREG

#define PIN(x)					\
do {						\
    status = (x);					\
    if ((status) != 0)				\
        return status;				\
} while (0)

int hal_setup(int id, haldata_t *h, const char *name)
{
    int status;
    PIN(hal_pin_bit_newf(HAL_OUT, &(h->e_stop), id, "%s.e-stop", name));
    PIN(hal_pin_bit_newf(HAL_OUT, &(h->stop), id, "%s.stop", name));
    PIN(hal_pin_bit_newf(HAL_OUT, &(h->run), id, "%s.run", name));
    PIN(hal_pin_u32_newf(HAL_OUT, &(h->mode), id, "%s.mode", name));
    PIN(hal_pin_s32_newf(HAL_OUT, &(h->feed_override), id, "%s.feed-override", name));
    PIN(hal_pin_s32_newf(HAL_OUT, &(h->speed_override), id, "%s.speed-override", name));
    PIN(hal_pin_u32_newf(HAL_OUT, &(h->first_key), id, "%s.first_key", name));
    PIN(hal_pin_u32_newf(HAL_OUT, &(h->second_key), id, "%s.second_key", name));
    PIN(hal_pin_s32_newf(HAL_OUT, &(h->errorcount), id, "%s.error-count", name));
    
    PIN(hal_pin_bit_newf(HAL_IN, &(h->x_is_home), id, "%s.x-is-home", name));
    PIN(hal_pin_bit_newf(HAL_IN, &(h->y_is_home), id, "%s.y-is-home", name));
    PIN(hal_pin_bit_newf(HAL_IN, &(h->z_is_home), id, "%s.z-is-home", name));
    
    return 0;
}
#undef PIN

int set_defaults(param_pointer p)
{
    haldata_t *h = p->haldata;
    
    *(h->e_stop) = 0;
    *(h->stop) = 0;
    *(h->run) = 0;
    *(h->mode) = 0;
    *(h->feed_override) = 0;
    *(h->speed_override) = 0;
    
    *(h->x_is_home) = 0;
    *(h->y_is_home) = 0;
    *(h->z_is_home) = 0;
    
    p->failed_reg = 0;
    return 0;
}

int main(int argc, char **argv)
{
    int opt;
    param_pointer p = &param;
    int retval = -1;

    uint16_t tmp = 0x0204;
    const uint16_t *speed = &tmp;
        
    p->progname = argv[0];
    connection_state = NOT_CONNECTED;
    
    while ((opt = getopt_long(argc, argv, option_string, long_options, NULL)) != -1) {
        switch(opt) {
            case 'm':
                p->modbus_debug = 1;
                break;
            case 'd':
                p->debug = 1;
                break;
            case 'P':
                p->device = optarg;
                break;
            case 'h':
            default:
                usage(argc, argv);
                exit(0);
        }
    }
    
    
    signal(SIGINT, quit);
    signal(SIGTERM, quit);
    signal(SIGUSR1, toggle_debug);
    signal(SIGUSR2, toggle_modbus_debug);
    
    // create HAL component 
    p->hal_comp_id = hal_init(p->modname);
    if ((p->hal_comp_id < 0) || (connection_state == DONE)) {
        fprintf(stderr, "%s: ERROR: hal_init(%s) failed: HAL error code=%d\n",
                p->progname, p->modname, p->hal_comp_id);
        retval = p->hal_comp_id;
        goto finish;
    }
    
    // grab some shmem to store the HAL data in
    p->haldata = (haldata_t *)hal_malloc(sizeof(haldata_t));
    if ((p->haldata == 0) || (connection_state == DONE)) {
        fprintf(stderr, "%s: ERROR: unable to allocate shared memory\n", p->modname);
        retval = -1;
        goto finish;
    }
    if (hal_setup(p->hal_comp_id,p->haldata, p->modname))
        goto finish;
    
    set_defaults(p);
    hal_ready(p->hal_comp_id);
    
    DBG("using libmodbus version %s\n", LIBMODBUS_VERSION_STRING);
    
    connection_state = OPENING;
    if ((p->ctx = modbus_new_rtu(p->device, p->baud, p->parity, p->bits, p->stopbits)) == NULL) {
        fprintf(stderr, "%s: ERROR: modbus_new_rtu(%s): %s\n",
                p->progname, p->device, modbus_strerror(errno));
        goto finish;
    }
    DBG("device(%s) baud(%d) parity(%s) bits(%d) stopbits(%d)\n", p->device, p->baud, &(p->parity), p->bits, p->stopbits);
    if (modbus_set_slave(p->ctx, p->slave) < 0) {
        fprintf(stderr, "%s: ERROR: invalid slave number: %d\n", p->modname, p->slave);
        goto finish;
    }
    if ((retval = modbus_connect(p->ctx)) != 0) {
        fprintf(stderr, "%s: ERROR: couldn't open serial device: %s\n", p->modname, modbus_strerror(errno));
        goto finish;
    }
    DBG("%s: serial port %s connected at %s\n", p->progname, p->device, p->baud);
    
    modbus_write_registers(p->ctx, BAUD_REG, 1, speed);

    modbus_flush(p->ctx);
    modbus_close(p->ctx);

    p->baud = 57600;
    
        connection_state = OPENING;
    if ((p->ctx = modbus_new_rtu(p->device, p->baud, p->parity, p->bits, p->stopbits)) == NULL) {
        fprintf(stderr, "%s: ERROR: modbus_new_rtu(%s): %s\n",
                p->progname, p->device, modbus_strerror(errno));
        goto finish;
    }
    DBG("device(%s) baud(%d) parity(%s) bits(%d) stopbits(%d)\n", p->device, p->baud, &(p->parity), p->bits, p->stopbits);
    if (modbus_set_slave(p->ctx, p->slave) < 0) {
        fprintf(stderr, "%s: ERROR: invalid slave number: %d\n", p->modname, p->slave);
        goto finish;
    }
    if ((retval = modbus_connect(p->ctx)) != 0) {
        fprintf(stderr, "%s: ERROR: couldn't open serial device: %s\n", p->modname, modbus_strerror(errno));
        goto finish;
    }
    DBG("%s: serial port %s connected at %s\n", p->progname, p->device, p->baud);

    
    connection_state = CONNECTED;

    while (connection_state != DONE) {
        
        while (connection_state == CONNECTED) {
            if ((retval = read_data(p->ctx, p->haldata, p))) {
                p->modbus_ok = 0;
            } else {
                p->modbus_ok = 1;
            }
            
            if ((retval = write_data(p->ctx, p->haldata, p))) {
                p->modbus_ok = 0;
                if ((retval == EBADF || retval == ECONNRESET || retval == EPIPE)) {
                    connection_state = RECOVER;
                }
            } else {
                p->modbus_ok++;
            }
            
        }
        
        switch (connection_state) {
            case DONE:
                // cleanup actions before exiting.
                modbus_flush(p->ctx);
                break;
                
            case RECOVER:
                DBG("recover\n");
                set_defaults(p);
                p->read_initial_done = 0;
                // reestablish connection to slave
                
                modbus_flush(p->ctx);
                modbus_close(p->ctx);
                while ((connection_state != CONNECTED) && (connection_state != DONE)) {
                    sleep(p->reconnect_delay);
                    if (!modbus_connect(p->ctx)) {
                        connection_state = CONNECTED;
                        DBG("rtuclient reconnect\n");
                    } else {
                        fprintf(stderr, "%s: recovery: modbus_connect(): %s\n",
                        p->progname, modbus_strerror(errno));
                    }
                }
                    
                    break;
            default: ;
        }
    }
    
    finish:
    windup(p);
    return retval;
}

